<?php

/**
 * Class OcConnectionStatic
 * Handles the background update queue of json data.
 *
 */
class OcConnectionStatic {

    public static $exec_chain = array();
    public static $key;
    protected static $_instance;

    public static function getInstance() {
        if ( null === self::$_instance ) {
            $key = 'cache_update_queue_';
            $key .= substr( str_shuffle( '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ' ), 0, 10 );

            self::$key = $key;
            self::$_instance = new self();
        }

        return self::$_instance;
    }

    public function __construct() {
    }

    /**
     * On destruct execute cli command to update needed cases.
     */
    public function __destruct() {

        if ( count( self::$exec_chain ) > 0 ) {
            $cmd = PHP_BIN . ' ' . __DIR__ . '/cache_update.php ' . self::$key . ' ' . $_SERVER['HTTP_HOST'] . ' > /dev/null &';
            exec( $cmd );
        }
    }
}

/**
 * Class OcConnection
 *
 * Class to enable communication with Open Content
 *
 * Use wrapping OcAPI class to interact with OC
 *
 * @author Infomaker Scandinavie AB
 *
 */
class OcConnection {
    
    const TRANSIENT_KEY = 'ew_t';
    const TRANSIENT_EXPIRE_TIME = PHP_OB_CACHE_TTL;
    
    /**
     * @var int
     */
    protected $latestQueryTotalHits;
    
    /**
     * @var array
     */
    private $remote_args;
    
    /**
     * @var string
     */
    protected $oc_url;
    
    /**
     * @var string
     */
    private $oc_base_url;
    
    /**
     * @var array
     */
    private $return_array = array();
    
    /**
     * @var OpenContent
     */
    private $open_content;
    
    /**
     * @var array
     */
    private $exec_chain = array();

    /**
     * Public getter for the OC Base URL
     *
     * @scope public
     *
     * @return null|string
     */
    public function getOcBaseUrl() {
        return isset( $this->oc_base_url ) ? $this->oc_base_url : null;
    }

    /**
     * Public getter for the full OC URL
     *
     * @scope public
     *
     * @return null|string
     */
    public function getOcUrl() {
        return isset( $this->oc_url ) ? $this->oc_url : null;
    }

    /**
     * Public setter for the OC URL
     *
     * @scope public
     *
     * @param string $url
     */
    public function setOcUrl( $url ) {
        $this->oc_url = $url;
    }

    /**
     * Public getter for the OC Auth
     *
     * @scope public
     *
     * @return null
     */
    public function getOcAuth() {
        return isset( $this->oc_auth ) ? $this->oc_auth : null;
    }

    /**
     * Public getter for cache TTL
     *
     * @scope public
     *
     * @return int
     */
    public function getCacheTTL() {
        return $this->open_content->getTimeToCacheJson();
    }
    
    /**
     * Public getter to get the latest queries total hits
     *
     * @since 0.1
     * @return int
     */
    public function getLatestQueryTotalHits() {
        return $this->latestQueryTotalHits;
    }

    /**
     * Public getter for remote args.
     * @return array
     */
    public function getRemoteArgs() {
        return $this->remote_args;
    }

    /**
     * Public constructor
     *
     * @scope public
     *
     * @param OpenContent $oc
     * @param string $query
     * @param string $start
     * @param string $limit
     * @param string $sort
     */
    public function __construct( OpenContent $oc, $query = '', $start = '', $limit = '', $sort = '' ) {
        if ( !empty( $query ) ) {
            trigger_error( 'Oc_Connection is being used in a deprecated way, use OcApi instead', E_STRICT );
        }
        $this->open_content = $oc;
        $this->oc_base_url = $oc->getOcBaseUrl();
        $this->remote_args = array(
            'headers' => array(
                'Authorization' => 'Basic ' . base64_encode( $oc->getOcUserName() . ':' . $oc->getOcPassword() )
            ),
            'timeout' => 20
        );
        $this->oc_url = $this->prep_oc_query( $query, $start, $limit, $sort );

        OcConnectionStatic::getInstance();
    }

    /**
     * Private function called to sanitize and prepare the given query
     *
     * @scope private
     *
     * @param string $query
     * @param string $start
     * @param string $limit
     * @param string $sort
     *
     * @return null|string
     */
    protected function prep_oc_query( $query = '', $start = '', $limit = '', $sort = '' ) {
        $prepared_query = $this->_build_query_string( $query, $start, $limit, $sort, $url_encoded = true );
        $prepared_query = $this->append_search_properties( $prepared_query );
        if ( $prepared_query !== '' ) {
            return $this->open_content->getOcBaseUrl() . "search?{$prepared_query}";
        }
        
        return $this->open_content->getOcBaseUrl();
    }

    /**
     * Function to prepare a Suggest query
     *
     * @scope public
     *
     * @param string $selector
     * @param string $query
     * @param string $start
     * @param int $limit
     *
     * @return void
     */
    public function prepare_oc_suggest_query( $selector, $query, $start = '', $limit = 25 ) {
        if ( null !== $selector ) {
            $this->oc_url = $this->open_content->getOcBaseUrl() . "suggest?field={$selector}";
        }

        $prepared_query = $this->_build_query_string( $query, $start, $limit, $sort = '', $url_encode = true );

        if ( $prepared_query !== '' ) {
            $this->oc_url .= "&{$prepared_query}";
        }
    }

    /**
     * Function to build query string
     *
     * @scope    protected
     *
     * @param $query
     * @param string $start
     * @param string $limit
     * @param string $sort
     * @param bool $url_encoded
     *
     * @return string
     */
    protected function _build_query_string( $query, $start, $limit, $sort, $url_encoded = false ) {

        if ( get_option( 'timezone_string' ) ) {
            date_default_timezone_set( get_option( 'timezone_string' ) );
        }

        $prepared_query = '';

        if ( $query !== '' ) {

            if ( class_exists( 'EnvSettings' ) ) {
                $env_settings = New EnvSettings();
                $env_query = $env_settings->get_env_query();

                if ( $env_query !== '' ) {
                    $query .= ' ' . trim( $env_query );
                }
            }

            $first_char = $query[ 0 ];

            if ( $first_char === '/' ) {
                $query = substr( $query, 1 );
            }

            $url_start = '';
            if ( null !== $start && $start !== '' ) {
                $url_start = "&start={$start}";
            }

            $url_limit = '';
            if ( null !== $limit && $limit !== '' ) {
                $url_limit = "&limit={$limit}";
            }

            $url_sort = '';
            if ( null !== $sort && $sort !== '' ) {
                $url_sort = "&sort.name={$sort}";
            }

            /*
             * Shortcode Param replace segment
             */
            if ( isset( $_GET[ 'date' ] ) ) {
                $date = wp_filter_nohtml_kses( $_GET[ 'date' ] );
            } else {
                $date = gmdate( 'Y-m-d' );
            }
            $gm_start_date = gmdate( "Y-m-d\TH:i:s\Z", strtotime( $date . '00:00:00' ) );
            $gm_end__date = gmdate( "Y-m-d\TH:i:s\Z", strtotime( $date . '23:59:59' ) );

            $section = '* OR (*:* -ShopSection:*)';
            if ( isset( $_GET[ 'section' ] ) ) {
                $section = '"' . wp_filter_nohtml_kses( $_GET[ 'section' ] ) . '"';
            }
            
            $tags = '* OR (*:* -Tags:*)';
            if ( isset( $_GET[ 'tags' ] ) ) {
                $tags = '"' . wp_filter_nohtml_kses( $_GET[ 'tags' ] ) . '"';
            }
            
            $author = '* OR (*:* -Author:*)';
            if ( isset( $_GET[ 'author' ] ) ) {
                $author = '"' . wp_filter_nohtml_kses( $_GET[ 'author' ] ) . '"';
            }

            $text = '*';
            if ( isset( $_GET[ 'q' ] ) ) {
                $text = '' . wp_filter_nohtml_kses( $_GET[ 'q' ] ) . '';
            } else if ( isset( $_GET[ 'text' ] ) ) {
                $text = '' . wp_filter_nohtml_kses( $_GET[ 'text' ] ) . '';
            }

            $date_from = '';
            $date_to = '';
            if ( isset( $_GET[ 'date_from' ], $_GET[ 'date_to' ] ) ) {
                $date_from = wp_filter_nohtml_kses( $_GET[ 'date_from' ] );
                $date_to = wp_filter_nohtml_kses( $_GET[ 'date_to' ] );
            } else {
//                $field_pos = strpos( $query, '[param_date_interval]' );
                // TODO: Fix field mapping here so that updated is not hardcoded
                $query = preg_replace( "/\"?\[param_date_interval\]\"?/i", '* OR (*:* -Pubdate:*)', $query );
            }

            global $oc_widget_inject_query;
            $inject_query = '';
            if ( null !== $oc_widget_inject_query ) {
                $inject_query = $oc_widget_inject_query;
            }

            global $oc_api_base_query;
            $inject_base_query = '';
            if ( null !== $oc_api_base_query ) {
                $inject_base_query = $oc_api_base_query;
            }

            // If a global parameter for start is set, use it instead.
            global $oc_widget_inject_start;
            if ( null !== $oc_widget_inject_start ) {
                $url_start = "&start={$oc_widget_inject_start}";
            }

            //Shortcode replace
            $query = preg_replace( "/\"?\[param_date\]\"?/i", "[{$gm_start_date} TO {$gm_end__date}]", $query );
            $query = preg_replace( "/\"?\[param_date_interval\]\"?/i", "[{$date_from}T00:00:00Z TO {$date_to}T23:59:59Z]", $query );
            $query = preg_replace( "/\"?\[param_section\]\"?/i", $section, $query );
            $query = preg_replace( "/\"?\[param_tags\]\"?/i", $tags, $query );
            $query = preg_replace( "/\"?\[param_author\]\"?/i", $author, $query );
            $query = preg_replace( "/\"?\[param_text\]\"?/i", $text, $query );
            $query = preg_replace( "/\"?\[param_inject_query\]\"?/i", $inject_query, $query );
            $query = preg_replace( "/\"?\[oc_api_base_query\]\"?/i", $inject_base_query, $query );

            if ( $url_encoded ) {
                $query = urlencode( $query );
            }

            $prepared_query = "q={$query}" . $url_start . $url_limit . $url_sort;
        }

        return $prepared_query;
    }

    /**
     * Function to get remote content, using build in WP remote_get functions
     *
     * @scope protected
     *
     * @param string $url
     *
     * @return null|string
     */
    protected function _get_remote_content( $url ) {

        $start = microtime( true );
        
        $api_response = wp_remote_get( $url, $this->remote_args );
        if ( !is_wp_error( $api_response ) ) {
            $http_status = (int)wp_remote_retrieve_response_code( $api_response );
    
            if ( $http_status === 200 ) {
                $json_content = wp_remote_retrieve_body( $api_response );

                $debug = $this->get_debug_data_for_analyzer( $start );
                do_action( 'analyzer_add_data', [
                    'name' => 'Call from Oc_Connection',
                    'debug' => $debug
                ], 'OC' );

                return $json_content;
            }
            if ( $http_status === 401 ) {

                $debug = $this->get_debug_data_for_analyzer( $start );
                do_action( 'analyzer_add_data', [
                    'name' => 'Call from Oc_Connection with 401 response code',
                    'debug' => $debug
                ], 'OC_ERROR' );
                return '401';
            }
        }

        return null;
    }
    
    /**
     * @param int $start
     * @param int $backtrace
     *
     * @return array
     */
    protected function get_debug_data_for_analyzer( $start = null, $backtrace = 20 ) {
        if ( null === $start ) {
            $start = microtime( true );
        }

        $data = debug_backtrace( DEBUG_BACKTRACE_PROVIDE_OBJECT, $backtrace );
        $data[1]['time'] = [ 0 => microtime( true ) - $start ];
        return $data;

    }

    /**
     * function to get remote data, using build in WP remote_get functions
     *
     * @scope protected
     *
     * @param string $url
     *
     * @return null|string
     */
    protected function _get_remote_data( $url ) {
        $start = microtime( true );

        $api_response = wp_remote_get( $url, $this->remote_args );

        if ( !is_wp_error( $api_response ) ) {
            $http_status = (int)wp_remote_retrieve_response_code( $api_response );

            if ( $http_status === 200 ) {
                $debug = $this->get_debug_data_for_analyzer( $start );
                do_action( 'analyzer_add_data', [
                    'name' => 'Call from Oc_Connection',
                    'debug' => $debug
                ], 'OC' );
                return $api_response;
            }
            if ( $http_status === 401 ) {
                $debug = $this->get_debug_data_for_analyzer( $start );
                do_action( 'analyzer_add_data', [
                    'name' => 'Call from Oc_Connection 401 data',
                    'debug' => $debug
                ], 'OC_ERROR' );
                return '401';
            }
        }

        return null;
    }

    /**
     * Function to get json results from cache, if exists, else from Open Content and then store in cache
     *
     * @scope protected
     *
     * @param string $url
     * @param int $cache_ttl
     *
     * @return bool|null|string
     */
    protected function get_json_result( $url = null, $cache_ttl = null ) {

        $transient_ttl = $this->open_content->getTimeToCacheJson();
        if( $cache_ttl !== null && is_int( $cache_ttl ) ) {
            $transient_ttl = $cache_ttl;
        }

        if ( $url === null ) {
            $url = $this->oc_url;
        }

        $key = md5( $url );

        // Try to get content from cached transient.
        $result = get_transient( self::TRANSIENT_KEY . '_' . $key );
        $tmp_result = json_decode( $result );

        if ( $tmp_result === null || $tmp_result === false ) { // If no content is found in transient, get it directly from Open Content.
            $result = $this->update_transient_cache( $url, $key, $cache_ttl );
            $tmp_result = json_decode( $result );
        } else if ( false === get_transient( self::TRANSIENT_KEY . '_' . $key . '_expire' ) ) { // If timeout for query has occured, fill it up with async script for next user.
            OcConnectionStatic::$exec_chain[] = [ $url, $key ];
            set_transient( OcConnectionStatic::$key, OcConnectionStatic::$exec_chain, self::TRANSIENT_EXPIRE_TIME );
            set_transient( self::TRANSIENT_KEY . '_' . $key . '_expire', 'true', $transient_ttl );
        }

        if ( isset( $tmp_result->uuid ) ) {
            OcCacheHelper::getInstance()->add_to_cache_map( $tmp_result->uuid, $key, self::TRANSIENT_EXPIRE_TIME );
        } else if ( isset( $tmp_result->contentType ) && strtolower( $tmp_result->contentType ) === 'article' ) {
            $uuid = '';
            if( isset( $tmp_result->properties ) && is_array( $tmp_result->properties ) ) {
                foreach ( $tmp_result->properties as $prop ) {
                    if( $prop->name === 'uuid' ) {
                        $uuid = $prop->values[ 0 ];
                        break;
                    }
                }
            }

            OcCacheHelper::getInstance()->add_to_cache_map( $uuid, $key, self::TRANSIENT_EXPIRE_TIME );
        } else if ( isset( $tmp_result->hits, $tmp_result->hits->hits ) && is_array( $tmp_result->hits->hits ) ) { //Multiple articles
            foreach ( $tmp_result->hits->hits as $hit ) {
                OcCacheHelper::getInstance()->add_to_cache_map( $hit->id, $key, self::TRANSIENT_EXPIRE_TIME );
            }
        }

        return $result;
    }
    
    /**
     * Function to update a transients cache with content from Open Content.
     *
     * @param string $url
     * @param string $key
     * @param null|int $cache_ttl
     *
     * @return null|string
     */
    public function update_transient_cache( $url, $key, $cache_ttl = null ) {
        $result = $this->_get_remote_content( $url );
        $this->set_transient_cache( $result, $key, $cache_ttl );
        return $result;
    }
    
    /**
     * Helper function to set transient
     *
     * @param string $to_cache
     * @param string $key
     * @param null|int $cache_ttl
     *
     * @return void
     */
    private function set_transient_cache( $to_cache, $key, $cache_ttl = null ) {

        $transient_ttl = $this->open_content->getTimeToCacheJson();
        if( $cache_ttl !== null && is_int( $cache_ttl ) ) {
            $transient_ttl = $cache_ttl;
        }

        if ( $to_cache !== null && $to_cache !== '401' ) {
            set_transient( self::TRANSIENT_KEY . '_' . $key, $to_cache, self::TRANSIENT_EXPIRE_TIME );
            set_transient( self::TRANSIENT_KEY . '_' . $key . '_expire', 'true', $transient_ttl );
        } else {
            set_transient( self::TRANSIENT_KEY . '_' . $key . '_expire', 'true', 120 );
        }
    }
    
    /**
     * @param array $response
     *
     * @return void
     */
    public function update_transient_notifier_push( $response ) {
        //check all keys that the uuid resides
        $transient_keys = OcCacheHelper::getInstance()->get_transient_keys_by_uuid( $response['uuid'] );
        $response_json = json_decode( $response['response'] );

        if ( count( $transient_keys ) > 0 ) {
            foreach ( $transient_keys as $transient_key => $expire_time ) {
                $json = json_decode( get_transient( self::TRANSIENT_KEY . '_' . $transient_key ), true );

                if ( isset( $json['hits'] ) ) {

                    foreach ( (array)$json['hits']['hits'] as $hit_key => $hit ) {

                        if ( isset( $hit['id'] ) || isset( $hit['uuid'] ) ) {
                            $uuid = isset( $hit['id'] ) ? $hit['id'] : $hit['uuid'];

                            if ( $uuid === $response['uuid'] ) {
                                $json['hits']['hits'][$hit_key]['versions'][0]['properties'] = $this->format_multi_properties( $response_json->properties );
                            }
                        }
                    }

                    $this->set_transient_cache( json_encode( $json, JSON_UNESCAPED_UNICODE ), $transient_key );
                    OcCacheHelper::getInstance()->update_cache_map( $response['uuid'], $transient_key, self::TRANSIENT_EXPIRE_TIME );
                } else if ( isset( $json['contentType'] ) && strtolower( $json['contentType'] ) === 'article' ) {
                    $json['properties'] = $this->format_single_properties( $response_json->properties );
                    $this->set_transient_cache( json_encode( $json, JSON_UNESCAPED_UNICODE ), $transient_key );
                    OcCacheHelper::getInstance()->update_cache_map( $response['uuid'], $transient_key, self::TRANSIENT_EXPIRE_TIME );
                }

            }
        }
    }
    
    /**
     * @param array $response
     *
     * @return void
     */
    public function add_transient_notifier_push( $response ) {
        $key = md5( $this->oc_base_url . "objects/{$response['uuid']}/properties" );

        $this->set_transient_cache( $response['response'], $key );
        OcCacheHelper::getInstance()->add_to_cache_map( $response['uuid'], $key, self::TRANSIENT_EXPIRE_TIME );
    }
    
    /**
     * @param string $uuid
     *
     * @return bool
     */
    public function delete_transient_notifier_push( $uuid ) {
        $ret = false;
        $transient_keys = OcCacheHelper::getInstance()->get_transient_keys_by_uuid( $uuid );

        if ( count( $transient_keys ) > 0 ) {
            $ret = true;
            foreach ( $transient_keys as $transient_key => $expire_time ) {
                $json = json_decode( get_transient( self::TRANSIENT_KEY . '_' . $transient_key ), true );

                if ( isset( $json['hits'] ) ) {
                    foreach ( (array)$json['hits']['hits'] as $hit_key => $hit ) {

                        if ( $hit['id'] === $uuid ) {
                            unset( $json['hits']['hits'][$hit_key] );
                            OcCacheHelper::getInstance()->remove_from_cache_map( $uuid, $transient_key );
                        }
                    }

                    $this->set_transient_cache( json_encode( $json, JSON_UNESCAPED_UNICODE ), $transient_key );
                } else if ( isset( $json['contentType'] ) && strtolower( $json['contentType'] ) === 'article' ) {

                    delete_transient( self::TRANSIENT_KEY . '_' . $transient_key );
                    delete_transient( self::TRANSIENT_KEY . '_' . $transient_key . '_expire' );
                    OcCacheHelper::getInstance()->remove_from_cache_map( $uuid, $transient_key );
                }
            }
        }

        return $ret;
    }
    
    /**
     * @param array $properties
     *
     * @return array
     */
    private function format_single_properties( array $properties = [] ) {
        $ret = array();

        foreach ( $properties as $prop ) {
            $ret[] = array(
                'name' => $prop->name,
                'values' => $prop->values
            );
        }

        return $ret;
    }
    
    /**
     * @param $properties
     *
     * @return array
     */
    private function format_multi_properties( array $properties = [] ) {
        $ret = array();

        foreach ( $properties as $prop ) {
            $ret[$prop->name] = $prop->values;
        }

        return $ret;
    }


    /**
     * Function to clear cache on Widget
     *
     * @scope public
     *
     * @param $md5_key
     * @return void
     */
    public static function clear_widget_cache( $md5_key=null ) {
        if ( null !== $md5_key ) {
            $result = delete_transient( self::TRANSIENT_KEY . '_' . $md5_key );
            do_action( 'clear_widget_cache_hook', $result, $md5_key );
        }
    }

    /**
     * Function to update Widget cache
     * This function is called when OC Notifier tells us there has
     * been a CRUD event on an Widget query result
     * OR when a widget is saved
     * This function will flush existing cache and get new JSON from OC and store it in transient
     *
     * @scope public
     *
     * @param bool $oc_query
     * @return void
     */
    public function update_widget_cache( $oc_query = false ) {
        if ( $oc_query === false ) {
            //Widget updated, take action...
            $url = $this->oc_url;

        } else {
            //Notifier calling, take action...
            $url = $this->oc_base_url . 'search?' . $oc_query;
        }

        $key = md5( $url );
        $this->update_transient_cache( $url, $key );
    }

    /**
     * DEPRECATED
     *
     * Function to be called by OC to invalidate cache on ALL Widgets
     * Will trigger widgets to empty their cache
     * @deprecated since 1.5.0 to be removed in 1.7.0
     */
    public function oc_event_clear_cache() {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. To be removed in 1.7.0', __METHOD__), E_USER_DEPRECATED);
        require_once '../../../../wp-load.php';

        $widgets = array();
        $oc_article_widgets = array();
        $oc_article_widgets_custom = array();

        //Get all page widgets (widgets from static pages like Home etc)
        $pages = get_pages();
        foreach ( $pages as $key => $page ) {
            if ( $key !== 'wp_inactive_widgets' ) {
                $widget    = get_post_meta( $page->ID, '_sidebars_widgets', true );
                $widgets[] = $widget;
            }
        }

        //Get all oc article widgets (General widgets set on Articles-widgets page)
        global $sidebars_widgets;
        if ( !empty( $sidebars_widgets ) ) {
            foreach ( $sidebars_widgets as $key => $widget ) {
                if ( $key !== 'wp_inactive_widgets' ) {
                    $oc_article_widgets[] = $widget;
                }
            }
        }

        //Get all custom oc article widgets (customized widgets on specific articles)
        $args = array( 'post_type' => 'article' );
        $oc_articles = get_posts( $args );
        foreach ( $oc_articles as $key => $article ) {
            if ( $key !== 'wp_inactive_widgets' ) {
                $oc_article_widgets_custom[] = get_post_meta( $article->ID, '_sidebars_widgets', true );
            }
        }

        $widgets[] = $oc_article_widgets;
        $widgets   = array_merge( $widgets, $oc_article_widgets_custom );

        foreach ( $widgets as $key => $widget_array ) {
            if ( $widget_array === '' ) {
                unset( $widgets[$key] );
            }

            if ( is_array( $widget_array ) && count( $widget_array ) > 0 ) {
                foreach ( $widget_array as $widget_area ) {
                    if ( is_array( $widget_area ) ) {
                        foreach ( $widget_area as $i => $widget_id ) {
                            self::clear_widget_cache( $widget_id ); //This line will clear the cache
                        }
                    }
                }
            }
        }
    }
    
    /**
     * Function to get an array with articles from Open Content
     *
     * @param bool $create_cpt
     *
     * @since 1.5.0
     * @return array
     */
    public function get_oc_articles( $create_cpt = true ) {
        $result = $this->get_json_result();
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            $this->latestQueryTotalHits = isset( $json_result->hits, $json_result->hits->totalHits ) ? $json_result->hits->totalHits : 0;
            try {
                foreach ( (array)$json_result->hits->hits as $object ) {
                    $this->return_array[] = $this->create_oc_search_object( $object->versions[0]->properties, $create_cpt );
                }
            } catch (Exception $e) {
                trigger_error( 'Error in get_oc_articles - Foreach json->hits->hits, Exception: ' . $e->getTraceAsString(), E_USER_ERROR );
            }
        }

        return $this->return_array;
    }

    /**
     * Function to get an array with ads from Open Content
     *
     * @scope public
     *
     * @return array
     */
    public function get_oc_ads() {
        $result = $this->get_json_result();
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            try {
                foreach ( (array)$json_result->hits->hits as $object ) {
                    $this->return_array[] = $this->create_oc_search_object( $object->versions[0]->properties );
                }
            } catch (Exception $e) {
                trigger_error( 'Error in get_oc_ads - Foreach json->hits->hits, Exception: ' . $e->getTraceAsString(), E_USER_ERROR );
            }

        }

        return $this->return_array;
    }

    /**
     * Function to get an array with Images from Open Content
     *
     * @scope public
     *
     * @return array
     */
    public function get_oc_images() {
        $result = $this->get_json_result();
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            $this->latestQueryTotalHits = ( isset( $json_result->hits ) && isset( $json_result->hits->totalHits ) ) ? $json_result->hits->totalHits : 0;

            try {
                foreach ( (array)$json_result->hits->hits as $object ) {
                    $this->return_array[] = $this->create_oc_search_object( $object->versions[0]->properties );
                }
            } catch (Exception $e) {
                trigger_error( 'Error in get_oc_images - Foreach json->hits->hits, Exception: ' . $e->getTraceAsString(), E_USER_ERROR );
            }

        }

        return $this->return_array;
    }

    /**
     * Function to get full article with related images.
     *
     * @param string $article_id
     * @param array $prop_arr
     *
     * @return array with an OcArticle and an Array with images
     */
    public function get_single_article( $article_id, $prop_arr = null ) {

        global $use_oc_cache;

        $query = $this->oc_base_url . "objects/{$article_id}/properties";
        
        // If property array is set, add it's values to the query.
        $this->oc_url = is_array($prop_arr) ? "{$query}?" . $this->create_property_query($prop_arr) : $query;
        
        if ( $use_oc_cache ) {
            $result = $this->get_json_result();
        } else {
            $result = $this->_get_remote_content( $this->oc_url );
        }

        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
    
            $article = $this->get_oc_property_object($json_result);
    
            $this->return_array['article'] = $article;
            
            //TODO: Do we need this hardcoded dependency to the property "ImageUuid"?
            if ( isset( $article->imageuuid ) && is_array( $article->imageuuid ) && count( $article->imageuuid ) > 0 ) {
                $article_image_uuids = $article->imageuuid;
            } else {
                $article_image_uuids = $this->_get_single_article_images( $article->uuid[0] );
            }

            $this->return_array['article_images'] = $article_image_uuids;
        }
        return $this->return_array;
    }
    
    /**
     * Factory-function for creating OcObjects from data fetched from Open Content properties
     *
     * @param $object
     *
     * @since 1.5.0
     * @return AbstractOcObject
     */
    protected function get_oc_property_object( $object ) {
        $oc_object = $this->get_oc_object_by_type( isset($object->contentType) ? $object->contentType :'' );
    
        foreach ( (array)$object->properties as $prop ) {
            $oc_object->set( $prop->name, $this->get_oc_property_object_values( $prop->values ) );
        }
        
        if( $oc_object instanceof OcArticle ) {
            $this->create_article_post_type( $oc_object );
        }
    
        return $oc_object;
    }
    
    /**
     * Extract property values from Object fetched from a property search
     *
     * @param array $prop_values
     *
     * @since 1.5.0
     * @return array
     */
    protected function get_oc_property_object_values( array $prop_values = [] ) {
        return array_map(function($prop_value) {
            return is_object($prop_value) ? $this->get_oc_property_object($prop_value) : $prop_value;
        }, $prop_values);
    }
    
    /**
     * Function to instantiate a new OcObject depending on the contenttype
     *
     * @param string $type
     *
     * @since 1.5.0
     * @return AbstractOcObject
     */
    protected function get_oc_object_by_type( $type='' ) {
        $article_types = array_map( 'strtolower', $this->get_contenttypes_considered_articles() );
        $lower_contenttype = strtolower( implode('', (array)$type));
    
        if( in_array( $lower_contenttype, $article_types, true ) ) {
            return new OcArticle();
        }
        
        if( $lower_contenttype === 'ad' ) {
            return new OcAd();
        }
        
        if( $lower_contenttype === 'image' ) {
            return new OcImage();
        }
        
        return new OcObject();
    }
    
    /**
     * Get object data from Open Content.
     *
     * @param string $uuid
     * @param array  $prop_arr
     * @param bool   $use_cache
     *
     * @since 1.5.0
     * @return null|string
     */
    public function get_single_object_data( $uuid, array $prop_arr = [], $use_cache = true ) {

        $query = $this->oc_base_url . "objects/$uuid/properties";

        // If property array is set, add it's values to the query.
        if ( count( $prop_arr ) > 0) {

            foreach( $prop_arr as $key => $prop ) {
                $query .= $key === 0 ? '?' : '&' . 'name=' . $prop;
            }
        }

        $this->oc_url = $query;

        if ( $use_cache ) {
            $result = $this->get_json_result();
        } else {
            $result = $this->_get_remote_content( $this->oc_url );
        }

        if ( $result !== null && $result !== '401' ) {
            return $result;
        }

        return null;
    }

    /**
     * Function to set properties from OC property map
     *
     * @param $object
     *
     * @return void
     */
    public function add_mapped_properties( $object ) {
        $prop_map = OcUtilities::get_property_map();

        foreach ( $prop_map as $key => $value ) {
            $key = strtolower( $key );
            $temp = $object->$value;
            $object->$key = $temp;
        }
    }

    /**
     * Function to get all image metadata
     *
     * @scope public
     *
     * @param string $image_id
     * @param array $prop_arr
     *
     * @since 1.5.0
     * @return OcImage
     */
    public function get_single_image_metadata( $image_id, array $prop_arr = [] ) {

        $query = $this->oc_base_url . "objects/$image_id/properties";
        // If property array is set, add it's values to the query.
        if ( ! empty( $prop_arr ) ) {
            $c = 0;
            foreach ( $prop_arr as $prop ) {

                if ( $c === 0 ) {
                    $query .= '?name=' . $prop;
                } else {
                    $query .= '&name=' . $prop;
                }

                $c++;
            }
        }

        $result = $this->get_json_result( $query );
        $image = new OcImage();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            
            if ( isset($json_result->properties) ) {
                foreach ( (array)$json_result->properties as $prop ) {
                    $name = strtolower( $prop->name );
                    $val = $prop->values;
                    $image->$name = $val;
                }
            }
        }
        return $image;
    }

    /**
     * Function to be called by get_single_article to get related images
     *
     * @param string $article_id
     *
     * @return array
     */
    private function _get_single_article_images( $article_id ) {
        $query = $this->oc_base_url . "objects/$article_id/relations";

        $result = $this->get_json_result( $query );
        $uuid_arr = array();

        if ( $result !== null && $result !== '401' ) {

            $json_result = json_decode( $result );
            if ( count( $json_result->relations ) > 0 ) {

                foreach ( (array)$json_result->relations as $key => $value ) {

                    if ( $value->targetContentType === 'Image' && isset( $value->uuids ) && count( $value->uuids ) > 0 ) {

                        foreach ( (array)$value->uuids as $uuid ) {
                            $uuid_arr[] = $uuid;
                        }
                    }
                }
            }
        }

        return $uuid_arr;
    }

    /**
     * Function called by partial view to render image-src
     *
     * @scope public
     *
     * @param int $image_uuid
     * @param string $image_size
     * @param bool $source_file
     *
     * @return null|string
     */
    public function get_image( $image_uuid = 0, $image_size = 'thumbnail', $source_file = false ) {
        if ( $image_uuid === 0 ) {
            return 'fail';
        }

        if ( !$source_file ) {

            $query = $this->oc_base_url . "objects/$image_uuid/files/$image_size";
        } else {

            $query = $this->oc_base_url . "objects/$image_uuid";
        }

        return $this->_get_remote_content( $query );
    }

    /**
     * Function called by article view to display image-ul
     *
     * @scope public
     *
     * @param int $image_uuid
     * @param string $image_size
     * @param bool $source_file
     *
     * @return string
     */
    public function get_image_url( $image_uuid = 0, $image_size = 'thumbnail', $source_file = false ) {
        if ( $image_uuid === 0 ) {
            return 'fail';
        }

        if ( !$source_file ) {

            $query = $this->oc_base_url . "objects/$image_uuid/files/$image_size";
        } else {

            $query = $this->oc_base_url . "objects/$image_uuid";
        }

        return $query;
    }
    
    /**
     * @param array $params
     *
     * @return string
     */
    private function build_search_query( array $params ) {
        $params = array_replace_recursive( [
            'q' => '',
            'start' => 0,
            'limit' => 15,
            'contenttypes' => [],
            'properties' => [],
            'sort.name' => null,
            'sort.indexfield' => null,
            'facet' => null,
            'facet.limit' => null,
            'facet.mincount' => null,
            'facet.date.mincount' => null,
        ], $params );

        // We create our own parameters from these arrays
        $properties = $this->get_search_properties($params['properties']);
        $contenttypes = (array)$params['contenttypes'];
        unset( $params[ 'properties' ], $params[ 'contenttypes' ] );
    
        $env_settings = new EnvSettings();
        if ( $env_settings->get_use_trashed() ) {
            $params['q'] .= ' AND NOT EveryTrashed:"' . OcUtilities::get_site_shortname() . '"';
        }

        // Remove q-parameter if no query has been made
        $params['q'] = $params['q'] !== '' ? $params['q'] : null;

        // Convert facet-parameter if set
        if ( isset( $params['facet'] ) ) {
            $params['facet'] = $params['facet'] ? 'true' : 'false';
        }

        $query_string = http_build_query( $params );

        // Add contenttypes if specified
        foreach ( $contenttypes as $contenttype ) {
            $query_string .= "&contenttype={$contenttype}";
        }

        // Add property-parameters if properties has been specified.
        if ( count( $properties ) > 0 ) {

            // Add contenttype to the properties as it is needed by the function.
            if ( !in_array( 'contenttype', $properties, true ) ) {
                $properties[] = 'contenttype';
            }

            // Add the rest of the properties
            foreach ( $properties as $property ) {
                $query_string .= "&property={$property}";
            }
        }

        return "{$this->oc_base_url}search?{$query_string}";
    }
    
    /**
     * Function to get free text search from OC
     *
     * @scope public
     *
     * @param string $search_text
     * @param array  $include_array
     * @param string $sort
     * @param int    $facet_limit
     * @param int    $facet_mincount
     * @param int|string $start
     * @param int|string $limit
     * @param bool   $create_article
     * @param array  $prop_arr
     * @param bool   $use_cache
     * @param null   $use_facet
     *
     * @return array
     */
    public function text_search( $search_text, $include_array, $sort = '', $facet_limit = null, $facet_mincount = null, $start = '', $limit = '', $create_article = true, $prop_arr = [], $use_cache = false, $use_facet = null ) {

        // Redirect to new search function
        return $this->search( [
            'q' => $search_text,
            'contenttypes' => $include_array,
            'sort.name' => $sort,
            'facet' => $use_facet,
            'facet.limit' => $facet_limit,
            'facet.mincount' => $facet_mincount,
            'facet.date.mincount' => $facet_mincount,
            'start' => $start,
            'limit' => $limit,
            'properties' => $prop_arr
        ], $use_cache, $create_article );
    }
    
    /**
     * New function for using search against Open Content
     *
     * @param array $params
     * @param bool $use_cache
     * @param bool $create_article
     * @param int $cache_ttl
     *
     * @return array|string
     */
    public function search( array $params = [], $use_cache = true, $create_article = true, $cache_ttl = null ) {
        $this->return_array = array();

        $query = $this->_encode_query( $this->build_search_query( $params ) );
        return $this->get_search_result( $query, $use_cache, $create_article, $cache_ttl );
    }
    
    /**
     * @param string $query
     * @param bool   $use_cache
     * @param bool   $create_article
     * @param int    $cache_ttl
     *
     * @since 1.5.0
     * @return array
     */
    private function get_search_result( $query, $use_cache, $create_article, $cache_ttl = null ) {
        $result = null;

        if ( $use_cache ) {
            $result = $this->get_json_result( $query, $cache_ttl );
        } else {
            $result = $this->_get_remote_content( $query );
        }

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            $this->return_array['duration'] = $json_result->stats->duration;
            $this->return_array['hits'] = $json_result->hits->totalHits;
            $this->return_array['facet'] = $json_result->facet;
            
            foreach ( (array)$json_result->hits->hits as $object ) {
                $this->return_array[] = $this->create_oc_search_object( $object->versions[0]->properties, $create_article );
            }
        }

        return $this->return_array;
    }

    /**
     * Function to get Sections from Open Content
     *
     * @return array
     */
    public function get_oc_suggest() {
        $result = $this->get_json_result();
        $return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            try {
                foreach ( (array)$json_result->facetFields[0]->terms as $section ) {
                    $return_array[] = $section;
                }
            } catch (Exception $e) {
                trigger_error( 'Error in get_oc_suggest - Foreach json_result->facetFields->terms, Exception: ' . $e->getTraceAsString(), E_USER_ERROR );
            }

        }

        return $return_array;
    }
    
    /**
     * @param array $prop_values
     *
     * @since 1.5.0
     * @return array
     */
    protected function get_oc_search_object_values( array $prop_values = [] ) {
        return array_map(function($prop_value) {
            return is_object($prop_value) ? $this->create_oc_search_object($prop_value) : $prop_value;
        }, $prop_values);
    }
    
    /**
     * Function to create OC-Objects from the search-response out of data from Open Content
     *
     * @param      $object
     * @param bool $create_cpt
     *
     * @since 1.5.0
     * @return AbstractOcObject
     */
    protected function create_oc_search_object( $object, $create_cpt = true ) {
        $oc_object = $this->get_oc_object_by_type( isset( $object->contenttype ) ? $object->contenttype : '' );
        foreach ( (array)$object as $name => $val ) {
            $oc_object->set($name, $this->get_oc_search_object_values( $val !== null ? $val : [] ) );
        }
        
        if( $oc_object instanceof OcArticle ) {
            $this->create_article_post_type($oc_object, $create_cpt);
        }
        
        return $oc_object;
    }
    
    /**
     * Function to create articles in Wordpress if they meet the requirements
     *
     * @param OcArticle $article
     * @param bool      $create_cpt
     *
     * @since 1.5.0
     * @return void
     */
    protected function create_article_post_type( OcArticle $article, $create_cpt = true ) {
        $article->set_mapped_properties();
        if( $create_cpt && $this->article_can_be_created( $article ) ) {
            // Create a new Custom post type from the article object
            new OcArticleCustomPostFactory( $article );
        }
    }
    
    /**
     * Function to create Article object
     *
     * @scope protected
     *
     * @param      $article
     * @param bool $create_cpt
     *
     * @deprecated since 1.5.0 Use create_oc_search_object
     * @return OcArticle
     */
    protected function create_article_object( $article, $create_cpt = true ) {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Use "create_oc_search_object" instead', __METHOD__), E_USER_DEPRECATED);
        $oc_article = new OcArticle();
        foreach ( (array)$article->versions[0]->properties as $prop => $val ) {
            $name = strtolower( $prop );
            $oc_article->$name = $val;
        }
        if ( !isset( $oc_article->uuid[0] ) ) {
            $oc_article->uuid = array( $article->id );
        }

        $this->add_mapped_properties( $oc_article );

        if ( !$this->is_article( $oc_article ) ) {
            $create_cpt = false;
            trigger_error( 'Trying to create article post-type out of non-article' );
        }

        //Create an instance of our custom post type if wanted
        if ( $create_cpt ) {
            new OcArticleCustomPostFactory( $oc_article );
        }

        return $oc_article;
    }

    /**
     * function to create Ad object
     *
     * @scope protected
     *
     * @param $ad
     *
     * @deprecated since 1.5.0 Use create_oc_search_object
     * @return OcAd
     */
    protected function create_ad_object( $ad ) {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Use "create_oc_search_object" instead', __METHOD__), E_USER_DEPRECATED);
        $oc_ad = new OcAd();

        foreach ( (array)$ad->versions[0]->properties as $prop => $val ) {
            $name = strtolower( $prop );
            $oc_ad->$name = $val;
        }
        if ( !isset( $oc_ad->uuid[0] ) ) {
            $oc_ad->uuid = array( $ad->id );
        }

        return $oc_ad;
    }

    /**
     * function to create Image object
     *
     * @scope protected
     *
     * @param $image
     *
     * @deprecated since 1.5.0 Use create_oc_search_object
     * @return OcImage
     */
    protected function create_image_object( $image ) {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Use "create_oc_search_object" instead', __METHOD__), E_USER_DEPRECATED);
        $oc_img = new OcImage();

        foreach ( (array)$image->versions[0]->properties as $prop => $val ) {
            $name = strtolower( $prop );
            $oc_img->$name = $val;
        }

        if ( !isset( $oc_img->uuid ) ) {
            $oc_img->uuid = array( $image->id );
        }

        $this->add_mapped_properties( $oc_img );

        return $oc_img;
    }

    /**
     * function to create Page object
     *
     * @scope protected
     *
     * @param $page
     *
     * @deprecated since 1.5.0 Use create_oc_search_object
     * @return OcPage
     */
    protected function create_page_object( $page ) {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Use "create_oc_search_object" instead', __METHOD__), E_USER_DEPRECATED);
        $oc_page = new OcPage();

        foreach ( (array)$page->versions[0]->properties as $prop => $val ) {
            $name = strtolower( $prop );
            $oc_page->$name = $val;
        }

        if ( !isset( $oc_page->uuid[0] ) ) {
            $oc_page->uuid = array( $page->id );
        }

        return $oc_page;
    }

    /**
     * function to create Oc object
     *
     * @scope protected
     *
     * @param $object
     *
     * @deprecated since 1.5.0 Use create_oc_search_object
     * @return OcObject
     */
    protected function create_oc_object( $object ) {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Use "create_oc_search_object" instead', __METHOD__), E_USER_DEPRECATED);
        $oc_object = new OcObject();

        foreach ( (array)$object->versions[0]->properties as $prop => $val ) {
            $name = strtolower( $prop );
            $oc_object->$name = $val;
        }

        if ( !isset( $oc_object->uuid[0] ) ) {
            $oc_object->uuid = array( $object->id );
        }

        return $oc_object;
    }

    /**
     * Function to get OC Sort options
     *
     * @scope public
     *
     * @return array|mixed
     */
    public function get_oc_sort_options() {

        $key = 'ew_sort_options';
        if ( false === ( $this->return_array = get_transient( $key ) ) ) {

            $query = $this->oc_base_url . 'sortings';
            $result = $this->_get_remote_content( $query );

            $this->return_array = array();
            if ( $result !== null && $result !== '401' ) {
                $this->return_array = json_decode( $result );
            }

            set_transient( $key, $this->return_array, 300 );
        }

        return $this->return_array;
    }

    /**
     * Function to test oc query with Ajax call
     * TODO: Use new filter when adding articles?
     * @scope public
     *
     * @return array
     */
    public function ajax_test_query() {
        $result = $this->_get_remote_content( $this->oc_url );
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {

            $json_result = json_decode( $result );
            $hits = $json_result->hits;

            //If we get a response from OC but no hits
            if ( $hits->totalHits === 0 ) {
                $this->return_array['No Matches'] = __( 'Given query returns no matching objects from Open Content' );
            } else {
                foreach ( (array)$hits->hits as $object ) {
                    switch (strtolower( $object->versions[0]->properties->contenttype[0] )) {
                        case 'ad':
                            !isset( $this->return_array['Ads'] ) ? $this->return_array['Ads'] = 0 : null;
                            $this->return_array['Ads']++;
                            break;
                        case 'article':
                            !isset( $this->return_array['Articles'] ) ? $this->return_array['Articles'] = 0 : null;
                            $this->return_array['Articles']++;
                            break;
                        case 'image':
                            !isset( $this->return_array['Images'] ) ? $this->return_array['Images'] = 0 : null;
                            $this->return_array['Images']++;
                            break;
                        case 'page':
                            !isset( $this->return_array['Pages'] ) ? $this->return_array['Pages'] = 0 : null;
                            $this->return_array['Pages']++;
                            break;
                        case 'video':
                            !isset( $this->return_array['Video'] ) ? $this->return_array['Video'] = 0 : null;
                            $this->return_array['Video']++;
                            break;
                        default:
                            !isset( $this->return_array['Others'] ) ? $this->return_array['Others'] = 0 : null;
                            $this->return_array['Others']++;
                            break;
                    }
                }
            }

        } elseif ( $result === '401' ) //Not authorized
        {
            $this->return_array['Authentication'] = __( 'Invalid Open Content credentials. Check Open Content Settings' );
        } else //No response from OC or Null object is returned
        {
            $this->return_array['No Matches'] = __( 'No Response' );
        }

        return $this->return_array;
    }

    /**
     * Function to get Content types with Ajax (uses indexfield as value).
     *
     * @scope public
     *
     * @return array
     */
    public function ajax_get_content_types() {
        $cache_key = $this->generate_cache_key('contenttype_properties');
    
        // Fetch from cache
        if( ( $oc_properties = get_transient( $cache_key ) ) !== false ) {
            return $oc_properties;
        }
        
        $query = $this->oc_base_url . 'contenttypes';
        $result = $this->_get_remote_content( $query );
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );

            foreach ( (array)$json_result->contentTypes as $content_type ) {
                $this->return_array[$content_type->name] = array();

                foreach ( (array)$content_type->properties as $property ) {
                    $this->return_array[$content_type->name][] = $property->indexFields;
                }
            }
            set_transient( $cache_key, $this->return_array, self::TRANSIENT_EXPIRE_TIME );
        } else if ( $result === '401' ) {
            $this->return_array['Authentication'] = __( 'Invalid Open Content credentials. Check "Settings - Open Content"' );
        } else {
            $this->return_array['No Matches'] = __( 'No Content types found in Open Content' );
        }

        return $this->return_array;
    }
    
    /**
     * Function to get Content types and use property name as value.
     *
     * @param array $types
     *
     * @since 1.5.0
     * @return array
     */
    public function get_content_types_properties( array $types = [] ) {
        $cache_key = $this->generate_cache_key('contenttype_properties_new');

        // Fetch from cache
        if( ( $oc_properties = get_transient( $cache_key ) ) !== false ) {
            return $oc_properties;
        }

        $query = $this->oc_base_url . 'contenttypes';
        $result = $this->_get_remote_content( $query );
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );
            foreach ( (array)$json_result->contentTypes as $content_type ) {
                $this->return_array[$content_type->name] = array();

                foreach ( (array)$content_type->properties as $property ) {

                    $addable = true;
                    if( ! empty( $types ) && !in_array( strtolower( $property->type ), $types, true ) ) {
                        $addable = false;
                    }

                    if( $addable ) {
                        $this->return_array[$content_type->name][] = $property->name;
                    }
                }
            }
            set_transient( $cache_key, $this->return_array, self::TRANSIENT_EXPIRE_TIME );
        } else if ( $result === '401' ) {
            $this->return_array['Authentication'] = __( 'Invalid Open Content credentials. Check "Settings - Open Content"' );
        } else {
            $this->return_array['No Matches'] = __( 'No Content types found in Open Content' );
        }

        return $this->return_array;
    }

    /**
     * Function to get Index fields from Open Content
     *
     * @deprecated 1.5.0
     * @return array
     */
    public function get_index_fields() {
        @trigger_error(sprintf('Calling "%s" is deprecated since 1.5.0. Not up to date with current Open Content version', __METHOD__), E_USER_DEPRECATED);
        $query = $this->oc_base_url . 'admin/indexfields';
        $result = $this->_get_remote_content( $query );
        $this->return_array = array();

        if ( $result !== null && $result !== '401' ) {
            $json_result = json_decode( $result );

            foreach ( (array)$json_result->indexFields as $index_field ) {
                //TODO: Copy paste mistake. Is this function in use?
                if ( !empty( $property->indexFields ) ) {
                    $this->return_array[] = $index_field->name;
                }
            }
        } else if ( $result === '401' ) {
            $this->return_array['Authentication'] = __( 'Could not fetch Index fields due to Invalid Open Content credentials. Check "Settings - Open Content"' );
        } else {
            $this->return_array['No Matches'] = __( 'No Index Fields found in Open Content' );
        }

        return $this->return_array;
    }

    /**
     * Function to get Suggest results with AJAX
     *
     * @scope public
     *
     * @param string $field
     * @param string $q
     * @param string $incompleteWord
     * @param string $incompleteWordInText
     *
     * @return null|string
     */
    public function ajax_get_suggest( $field, $q, $incompleteWord, $incompleteWordInText ) {

        $query_str = 'suggest/';
        if ( $field !== null ) {
            $query_str = add_query_arg( 'field', urlencode( $field ), $query_str );
        }

        // TODO: This one is causing all kind of problems. Check if it is needed.
        /*if ( $q !== null && $field !== null ) {
            $query_str = add_query_arg( 'q', urlencode( $field . ':' . $q ), $query_str );
        } else*/
        if ( $q !== null ) {
            $query_str = add_query_arg( 'q', urlencode( $q ), $query_str );
        }

        if ( $incompleteWord !== null ) {
            $query_str = add_query_arg( 'incompleteWord', urlencode( $incompleteWord ), $query_str );
        }

        if ( $incompleteWordInText !== null ) {
            $query_str = add_query_arg( 'incompleteWordInText', urlencode( $incompleteWordInText ), $query_str );
        }

        $this->oc_url = $this->open_content->getOcBaseUrl() . $query_str;

        $result = $this->get_json_result( $this->oc_url );
        if ( $result !== null && $result !== '401' ) {
            return $result;
        }

        die();
    }

    /**
     * Function to url encode parameter values in a URL string
     * TODO Can we make this function more efficient?
     *
     * @scope private
     *
     * @param string $query
     * @return string
     */
    private function _encode_query( $query ) {
        $startpos = stripos( $query, '?' );
        $base = substr( $query, 0, $startpos + 1 );

        $query_str = substr( $query, $startpos + 1 );
        $query_str = explode( '&', $query_str );

        //TODO What is happening here?
        foreach ( $query_str as &$value ) {
            $tmp = explode( '=', $value );

            if ( isset( $tmp[1] ) ) {
                $tmp[1] = urlencode( urldecode( $tmp[1] ) );
            }
            $value = implode( '=', $tmp );
        }

        $encoded_url = implode( '&', $query_str );
        $encoded_url = $base . $encoded_url;

        return $encoded_url;
    }

    /**
     * Function to get the health status of the bound OpenContent
     * @return mixed
     */
    public function get_oc_health() {
        $ch = curl_init();
        curl_setopt( $ch, CURLOPT_URL, $this->oc_base_url . 'health' );
        curl_setopt( $ch, CURLOPT_RETURNTRANSFER, 1 );
        curl_setopt( $ch, CURLOPT_TIMEOUT, 5 );
        curl_exec( $ch );
        $response = curl_getinfo( $ch );
        curl_close( $ch );

        return $response;
    }

    /**
     * Determine if the value can be considered an article
     *
     * @param $article
     *
     * @return bool
     */
    protected function is_article( $article ) {
        if ( $article instanceof OcArticle ) {
            $article_types = array_map( 'strtolower', $this->get_contenttypes_considered_articles() );
            return in_array( strtolower( $article->get_value( 'contenttype' ) ), $article_types, true );
        }
        return false;
    }
    
    /**
     * Add requested properties in the form of query-parameters
     *
     * @param string $query
     *
     * @since 1.5.0
     * @return string
     */
    protected function append_search_properties( $query ) {
        $properties = $this->get_search_properties();
        
        if(! empty( $properties ) ) {
            foreach ( $properties as $property ) {
                $query .= "&property={$property}";
            }
        }
        
        return $query;
    }
    
    /**
     * Retrieve properties to request from Open Content.
     *
     * @param array $properties
     *
     * @since 1.5.0
     * @return array
     */
    protected function get_search_properties(array $properties = []) {
        
        // Add default properties if no properties has been specified
        if( empty( $properties ) ) {
            $properties = array_filter( (array)apply_filters( 'ew_apply_default_search_properties', [] ) );
        }
    
        // Make sure required properties are applied if properties has been requested.
        if(! empty( $properties ) ) {
            $properties = $this->append_required_properties($properties);
        }
    
        return array_unique( $properties );
    }
    
    /**
     * Add required properties if properties has been specified
     *
     * @param array $properties
     *
     * @since 1.5.0
     * @return array
     */
    protected function append_required_properties(array $properties = []) {
        if( empty( $properties ) ) {
            return [];
        }
       
        $required_oc_properties = array_filter([
            'contenttype',
            'uuid',
            $this->get_required_permalink_property()
        ]);
        
        // Merge and remove duplicates
        $properties = array_unique( array_merge( $properties, $required_oc_properties ) );
        
        $required_properties = array_filter( (array)apply_filters( 'ew_apply_required_search_properties', $properties ) );
    
        return $required_properties;
    }
    
    /**
     * Fetch and map the property required for creating an articles permalink
     *
     * @since 1.5.0
     * @return string
     */
    protected function get_required_permalink_property() {
        static $required_property;
        
        if ( $required_property !== null ) {
            return $required_property;
        }
        
        $cache_key = $this->generate_cache_key('required_permalink_property');
        
        // Fetch from cache
        if( ( $required_property = get_transient( $cache_key ) ) !== false ) {
            return $required_property;
        }
        
        $required_property  = OpenContent::getInstance()->getSlugProperties()[ 0 ];
        $article_properties = $this->get_contenttype_properties( 'Article' );
        $property_exists = in_array( $required_property, $article_properties, true );
        
        // If no property was found and a property map exists
        if(! $property_exists && (( $property_map = get_option( 'prop_map' ) ) !== false)) {
            $mapped_property = isset($property_map[$required_property]) ? $property_map[$required_property] : 'unknown';
            
            // Make sure that a property is mapped and not the property we already have
            if ( $mapped_property !== 'unknown' && strtolower($mapped_property) !== strtolower($required_property)) {
                
                foreach ( $article_properties as $prop ) {
                    
                    // Get the real name of the property and require it
                    if ( strtolower($mapped_property) === strtolower($prop) ) {
                        $required_property = $prop;
                    }
                }
            }
        }
        
        // Save property and return it
        set_transient( $cache_key, $required_property, self::TRANSIENT_EXPIRE_TIME );
        return $required_property;
    }

    /**
     * Get the contenttypes that should be considered articles.
     *
     * @return array
     */
    protected function get_contenttypes_considered_articles() {

        static $article_contenttypes;

        if( $article_contenttypes !== null ) {
            return $article_contenttypes;
        }

        $article_contenttypes = array_map( 'strtolower', apply_filters( 'ew_contenttypes_considered_article', [ 'article' ] ) );
        return $article_contenttypes;
    }
    
    /**
     * Check if an OcArticle has what it takes for an article post type to be created in Wordpress
     *
     * @param OcArticle $article
     *
     * @since 1.5.0
     * @return bool
     */
    protected function article_can_be_created( OcArticle $article ) {
        $properties = $this->append_required_properties(['contenttype', 'uuid']);
        $article_properties = $article->get_all_properties();
        
        // Walk through all required properties to see if we have what it takes
        $missing_values = array_filter($properties, function($property) use ($article_properties) {
            return in_array( strtolower($property), $article_properties, true );
        });
        
        return empty($missing_values);
    }
    
    /**
     * Helper function fetch all properties on given contenttype
     *
     * @param string $contenttype
     * @param array  $types
     *
     * @since 1.5.0
     * @return array
     */
    public function get_contenttype_properties( $contenttype, array $types = [] ) {
        $oc_properties = $this->get_content_types_properties( $types );

        if( isset( $oc_properties[ $contenttype ] ) ) {
            return array_filter( $oc_properties[ $contenttype ] );
        }
    
        return [];
    }
    
    /**
     * Function to generate a transient key
     *
     * @param string $suffix
     *
     * @since 1.5.0
     * @return string
     */
    public function generate_cache_key( $suffix='' ) {
        return self::TRANSIENT_KEY . "_{$suffix}";
    }
    
    /**
     * Combine all properties and create a query string using the provided field name
     *
     * @param array  $properties
     * @param string $field
     *
     * @return string
     * @since 1.5.0
     */
    private function create_property_query( array $properties, $field = 'name' ) {
        return "{$field}=" . implode( "&{$field}=", $properties );
    }
}